﻿using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  ///     Pushes the stack depth onto the INTEGER stack (thereby increasing it!).
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  [StorableClass]
  public abstract class StackdepthExpression<T> : StatelessExpression {
    protected StackdepthExpression() { }
    [StorableConstructor]
    protected StackdepthExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return false;
    }

    protected void Eval(IPushStack<T> stack, IPushStack<long> integerStack, bool incremental = false) {
      var count = stack.Count;
      if (incremental) count += 1;
      integerStack.Push(count);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Integer,
    "INTEGER.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.")]
  public class IntegerStackdepthExpression : StackdepthExpression<long> {
    public IntegerStackdepthExpression() { }
    [StorableConstructor]
    protected IntegerStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerStack, interpreter.IntegerStack, true);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Float,
    "FLOAT.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class FloatStackdepthExpression : StackdepthExpression<double> {
    public FloatStackdepthExpression() { }
    [StorableConstructor]
    protected FloatStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Boolean,
    "BOOLEAN.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class BooleanStackdepthExpression : StackdepthExpression<bool> {
    public BooleanStackdepthExpression() { }
    [StorableConstructor]
    protected BooleanStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Name,
    "NAME.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class NameStackdepthExpression : StackdepthExpression<string> {
    public NameStackdepthExpression() { }
    [StorableConstructor]
    protected NameStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.NameStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Exec,
    "EXEC.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class ExecStackdepthExpression : StackdepthExpression<Expression> {
    public ExecStackdepthExpression() { }
    [StorableConstructor]
    protected ExecStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.ExecStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Code,
    "CODE.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class CodeStackdepthExpression : StackdepthExpression<Expression> {
    public CodeStackdepthExpression() { }
    [StorableConstructor]
    protected CodeStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CodeStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.Char,
    "CHAR.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class CharStackdepthExpression : StackdepthExpression<char> {
    public CharStackdepthExpression() { }
    [StorableConstructor]
    protected CharStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CharStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.String,
    "STRING.STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class StringStackdepthExpression : StackdepthExpression<string> {
    public StringStackdepthExpression() { }
    [StorableConstructor]
    protected StringStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class IntegerVectorStackdepthExpression : StackdepthExpression<IReadOnlyList<long>> {
    public IntegerVectorStackdepthExpression() { }
    [StorableConstructor]
    protected IntegerVectorStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class FloatVectorStackdepthExpression : StackdepthExpression<IReadOnlyList<double>> {
    public FloatVectorStackdepthExpression() { }
    [StorableConstructor]
    protected FloatVectorStackdepthExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class BooleanVectorStackdepthExpression : StackdepthExpression<IReadOnlyList<bool>> {
    public BooleanVectorStackdepthExpression() { }
    [StorableConstructor]
    protected BooleanVectorStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack, interpreter.IntegerStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].STACKDEPTH",
    "Pushes the stack depth onto the INTEGER stack.",
    StackTypes.Integer)]
  public class StringVectorStackdepthExpression : StackdepthExpression<IReadOnlyList<string>> {
    public StringVectorStackdepthExpression() { }
    [StorableConstructor]
    protected StringVectorStackdepthExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack, interpreter.IntegerStack);
    }
  }
}